/*
 *	(c) 1995,1996 - Mathijs Maassen PE1NTP@PI8DAZ.#TWE.NLD.EU or tyz@xs4all.nl
 *
 *	Changes:
 *		  July 1995 - Initial version
 *		 4 Aug 1995 - v1.0
 *						Some cleanups and small changes
 *						Released to FBB @ EU
 *		12 Aug 1995 - v1.1
 *						Fixed stupid bug in read_inf()
 *						Released to FBB @ EU
 *		11 Dec 1996 - v2.0
 *						Ported to Linux
 *						RETMAIL.LNG added
 *						Warning messages added
 *
 *	Used TAB width: 4
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef __MSDOS__
#include <dos.h>
#include <alloc.h>
#endif
#include <time.h>
#include <string.h>
#include <errno.h>
#include "fbb.h"

#define SECS_PER_DAY 86400L
#define UNKNOWN 1
#define EXPIRED 2
#define WARNING	3

#define VER_STR  "RETMAIL v2.0 - (c) 1995,1996 Mathijs Maassen PE1NTP @ PI8DAZ.#TWE.NLD.EU\n\n"
#ifdef __MSDOS__
#define MAIL_FMT "%sMAIL%ld\\M_%06ld.MES"
#else
#define MAIL_FMT "%smail%ld/m_%06ld.mes"
#endif

typedef struct {
	char call[7];
	unsigned long rec;
} inf;

inf **call_arr;
int unknown_days, known_days, warn_days, debug;
char system_dir[128], mail_dir[128], mail_in[128];

enum { LNG_SUBJECT, LNG_SUBJWARN, LNG_REASON_EXP, LNG_REASON_UNK, LNG_EXPIRED, LNG_UNKNOWN, LNG_WARNING, lng_lines };

char *lng[lng_lines+1];

void *do_malloc(size_t size)
{
	void *p;

	if ((p = malloc(size)) == NULL) {
		fprintf(stderr, "Out of memory (allocating %d bytes)\n", size);
		exit(1);
	}

	return (p);
}

long filesize(char *fname)
{
	struct stat sb;

	if (stat(fname, &sb)) {
		perror(fname);
		exit(1);
	}

	return (sb.st_size);
}

FILE *open_file(char *name, char *mode)
{
	FILE *fp;

	if ((fp = fopen(name, mode)) == NULL) {
		fprintf(stderr, "Can't open %s: %s\n", name, sys_errlist[errno]);
		exit(1);
	}

	return (fp);
}

void print_lastlogin(FILE * out_fp, long rec)
{
	FILE *inf_fp;
	info data;
	char fname[128], tmstr[30];

	if (rec < 0)
		return;

	strcpy(fname, system_dir);
	strcat(fname, "inf.sys");

	inf_fp = open_file(fname, "rb");
	fseek(inf_fp, rec, SEEK_SET);
	fread(&data, sizeof(data), 1, inf_fp);
	fclose(inf_fp);

	strftime(tmstr, sizeof(tmstr), "%a %d %b %Y, %H:%M:%S GMT", gmtime(&data.hcon));
	fputs(tmstr, out_fp);
}

void printstr(int line, FILE * fp, int lf, bullist *msgdata, long inf_rec)
{
	int i;

	for (i = 0; lng[line][i]; i++)
		if (lng[line][i] != '$')
			fputc(lng[line][i], fp);
		else
			switch (lng[line][++i]) {
			case 'D':
				fprintf(fp, "%d", known_days);
				break;
			case 'U':
				fputs(msgdata->desti, fp);
				break;
			case 'X':
				print_lastlogin(fp, inf_rec);
				break;
			case 'S':
				fputs(msgdata->titre, fp);
				break;
			case 'W':
				fputc('\n', fp);
				break;
			case '$':
				fputc('$', fp);
				break;
			default:
				fputs("[Unknown escape char]", fp);
			}
	if (lf)
		fputc('\n', fp);
}

void read_lng(void)
{
	FILE *fp;
	char *lngf = "retmail.lng";
	int fsz;
	static char *lngp;
	int line, pos;

	fsz = filesize(lngf);
	fp = open_file(lngf, "r");
	lngp = do_malloc(fsz + 1);	/* +1 in case file doesn't end with \n */
	lngp[fsz] = '\0'; /* Make sure last string is nil terminated */
	fread(lngp, fsz, 1, fp);
	fclose(fp);

	lng[0] = lngp;
	for (pos = 0, line = 1; pos < fsz && line <= lng_lines; pos++) {
		if (lngp[pos] == '\n') {
			lngp[pos] = '\0';
			lng[line++] = &lngp[pos + 1];
		}
	}
}

void read_init_srv(void)
{
	FILE *fp;
	char line[128];
	int line_num = 0;

	fp = open_file("init.srv", "r");

	while (line_num != 14 && fgets(line, sizeof(line), fp) != NULL) {
		if (line[0] == '#')
			continue;
		line[strlen(line) - 1] = '\0';
		switch (++line_num) {
		case 5:
			strcpy(system_dir, line);
			break;
		case 6:
			strcpy(mail_dir, line);
			break;
		case 14:
			strcpy(mail_in, line);
			break;
		}
	}
	fclose(fp);
}

void read_inf(void)
{
	FILE *fp;
	info data;
	char fname[128];
	int entry = 0;
	long inf_size;

	strcpy(fname, system_dir);
	strcat(fname, "inf.sys");

	fprintf(stdout, "Reading %s.. ", fname);

	fp = open_file(fname, "rb");

	if ((inf_size = filesize(fname) / sizeof(data)) == 0) {
		fprintf(stderr, "%s is empty!\n", fname);
		exit(1);
	}

	call_arr = (inf **) do_malloc((inf_size + 1) * sizeof(inf *));

	while (fread(&data, sizeof(data), 1, fp) == 1) {
		if (!data.indic.callsign[0])
			continue;
		call_arr[entry] = (inf *) do_malloc(sizeof(inf));
		strcpy(call_arr[entry]->call, data.indic.callsign);
		call_arr[entry]->rec = ftell(fp) - sizeof(data);
		entry++;
	}

	call_arr[entry] = NULL;
	fclose(fp);
	fprintf(stdout, "OK - %d (out of %ld) entries\n", entry, inf_size);
}


void return_mail(bullist * data, int why, long inf_rec)
{
	FILE *in, *out;
	char line[256], mail_file[128], home_bbs[7], *p, *path = NULL;
	int headers = 1, path_len = 0, line_len = 0, i;

	home_bbs[0] = '\0';
#ifndef DEBUG
	sprintf(mail_file, MAIL_FMT, mail_dir, data->numero % 10, data->numero);
#else
	strcpy(mail_file, "mail");
#endif

	in = open_file(mail_file, "r");

	fseek(in, 0L, SEEK_END);
	if (!ftell(in))
		ungetc('\n', in); /* dirty trick to return empty msg's.. probably not portable */
	else
		fseek(in, 0L, SEEK_SET);

	out = open_file(mail_in, "a");

	while (fgets(line, sizeof(line), in) != NULL) {
		if (headers == 1 && !strncmp(line, "R:", 2)) {
			if ((p = strchr(line, '@')) != NULL) {
				if (*(++p) == ':')
					p++;
				strncpy(home_bbs, p, 6);
				home_bbs[6] = '\0';
				if ((p = strpbrk(home_bbs, ".\t\n ")) != NULL)
					*p = '\0';
				if (path_len == 0) {
					path = do_malloc(78);
					strcpy(path, "Path: !");
					path_len = 78;
					line_len = 7;
				}
				i = strlen(home_bbs);
				line_len += i + 1;
				if (line_len > 77) {
					path_len += 78;
					if ((path = realloc(path, path_len)) == NULL) {
						fprintf(stderr, "Out of memory (re-allocating %d bytes)\n", path_len);
						exit(1);
					}
					strcat(path, "\n      !");
					line_len = i + 8;
				}
				strcat(path, home_bbs);
				strcat(path, "!");
			}
			else
				headers = 0;
		}
		else {
			if (headers == 1) {
				if (home_bbs[0])
					fprintf(out, "#\nSP %s @ %s\n", data->exped, home_bbs);
				else
					fprintf(out, "#\nSP %s\n", data->exped);
				switch (why) {
				case EXPIRED:
					printstr(LNG_SUBJECT, out, 0, data, inf_rec);
					printstr(LNG_REASON_EXP, out, 1, data, inf_rec);
					printstr(LNG_EXPIRED, out, 1, data, inf_rec);
					break;
				case UNKNOWN:
					printstr(LNG_SUBJECT, out, 0, data, inf_rec);
					printstr(LNG_REASON_UNK, out, 1, data, inf_rec);
					printstr(LNG_UNKNOWN, out, 1, data, inf_rec);
					break;
				case WARNING:
					printstr(LNG_SUBJWARN, out, 0, data, inf_rec);
					printstr(LNG_REASON_UNK, out, 1, data, inf_rec);
					printstr(LNG_WARNING, out, 1, data, inf_rec);
					goto breakloop;
				}
				fputs("-----------------------------------------------------------------------------\n", out);
				if (path_len) {
					fprintf(out, "%s\n\n", path);
					free(path);
				}
				else
					fputc('\n', out);
				headers = 0;
				if (line[0] != '\n')
					fputs(line, out);
			}
			else
				fputs(line, out);
		}
	}

breakloop:

	fputs("\n/EX\n", out);
	fflush(out);
	if (ferror(out)) {
		perror("Error writing to MAIL.IN");
		exit(1);
	}

	fclose(out);
	fclose(in);
}

void kill_msg(FILE * fp, bullist * data)
{

	if (debug) {
		fputs("Nothing killed (DEBUG)\n", stdout);
		return;
	}

	fseek(fp, (int) -sizeof(bullist), SEEK_CUR);

	data->status = 'K';

	if (fwrite(data, 1, sizeof(bullist), fp) != sizeof(bullist)) {
		perror("Writing to DIRMES.SYS");
		exit(1);
	}
#ifdef __TURBOC__
	fseek(fp, 0L, SEEK_CUR);	/* otherwise next fread() returns EOF.. why? */
#endif
}

void mark_msg(FILE * fp, bullist * data)
{
	if (debug) {
		fputs("Nothing marked (DEBUG)\n", stdout);
		return;
	}

	fseek(fp, (int) -sizeof(bullist), SEEK_CUR);

	strcpy(&data->bbsv[1], "RETM");

	if (fwrite(data, 1, sizeof(bullist), fp) != sizeof(bullist)) {
		perror("Writing to DIRMES.SYS");
		exit(1);
	}
#ifdef __TURBOC__
	fseek(fp, 0L, SEEK_CUR);	/* otherwise next fread() returns EOF.. why? */
#endif
}

void read_dirmes(void)
{
	FILE *fp;
	bullist data;
	int i;
	time_t now;
	char fname[128];

	strcpy(fname, system_dir);
	strcat(fname, "dirmes.sys");

	fprintf(stdout, "Reading %s..\n", fname);

	fp = open_file(fname, "r+b");

	time(&now);

	while (fread(&data, sizeof(data), 1, fp) == 1) {
		if (data.bbsv[0] == '\0' && data.type == 'P' && data.status == 'N') {
			for (i = 0; call_arr[i]; i++) {
				if (!strcmp(call_arr[i]->call, data.desti))
					break;
			}
			if (call_arr[i] == NULL) {	/* Unknown user */
				if (now - data.date > SECS_PER_DAY * unknown_days) {
					fprintf(stdout, "Unknown: To %-6s from %-6s\n", data.desti, data.exped);
					return_mail(&data, UNKNOWN, -1);
					kill_msg(fp, &data);
				} else if (now - data.date > SECS_PER_DAY * warn_days && strcmp(&data.bbsv[1], "RETM")) {
					fprintf(stdout, "Warning: To %-6s from %-6s\n", data.desti, data.exped);
					return_mail(&data, WARNING, -1);
					mark_msg(fp, &data);
				}
			}
			else {		/* Known user */
				if (now - data.date > SECS_PER_DAY * known_days) {
					fprintf(stdout, "Expired: To %-6s from %-6s\n", data.desti, data.exped);
					return_mail(&data, EXPIRED, call_arr[i]->rec);
					kill_msg(fp, &data);
				}
			}
		}
	}
	fclose(fp);
}

void usage(void)
{
	fputs("Usage: RETMAIL [OPTIONS]\n\n"
		"  RETMAIL takes 3 options:\n"
		"      The first option is the number of days after which a message\n"
		"      is returned when the user is unknown.   The second option is\n"
		"      the number of days if the user is known on this system. The 3th\n"
		"      option is the number of days after which a warning is sent for\n"
		"      unknown-user messages. This value should of course be lower as the\n"
		"      number of days after which an unknown-user message is returned.\n\n"
		"      An optional 4th argument turns test mode on. In test mode returned\n"
		"      messages are written to RETMAIL.DBG instead of MAIL.IN, and will not\n"
		"      be killed.\n\n"
		"  For example:  RETMAIL 7 60 1\n\n"
		"      Will return messages for unknown users after 7 days, expiring\n"
		"      messages after 60 days, and warning messages are sent after one day.\n", stdout);
	exit(1);
}

int main(int argc, char **argv)
{
	fputs(VER_STR, stdout);

	debug = (argc == 5);
	if ((!debug && argc != 4) || sscanf(argv[1], "%d", &unknown_days) != 1 || sscanf(argv[2], "%d", &known_days) != 1 || sscanf(argv[3], "%d", &warn_days) != 1)
		usage();
	read_init_srv();
	if (debug)
		strcpy(mail_in, "RETMAIL.DBG");
	read_inf();
	read_lng();
	read_dirmes();
	return (0);
}
